import * as d3 from 'd3';
import { NetlistRender } from '.';
import { globalSetting } from '../global';
import { LAYOUT_CONSTANT } from './layout';

export class InstantiationRender {
    /**
     * 
     * @param {d3.Selection} selection 
     * @param {NetlistRender} rootRender 
     */
    constructor(selection, rootRender) {
        this.parentSelection = selection;
        this.rootRender = rootRender;

        /**
         * @type {BasicD3DataItem[]}
         */
        this.data = []

        /**
         * @description id 到管理数据项的映射
         * @type {Map<string, BasicD3ManagmentItem[]>}
         */
        this.id2manager = rootRender.id2manager;
    }

    /**
     * @description 将 elknode 关于 例化模块 的数据添加为 d3 数据项目
     * @param {ElkNode} node 
     */
    addAsD3DataItem(node) {
        const nodeModule = this.rootRender.nameToModule.get(node.renderName);
        const view = nodeModule.view;

        const portnames = [];
        let inputCount = 0;
        let outputCount = 0;
        const instanceWidth = LAYOUT_CONSTANT.INSTANTIATION_WIDTH + 2 * LAYOUT_CONSTANT.PORT_INNER_PADDING;

        // 统计出两侧 port 字符串最长的
        let inputPortMaxLength = 0;
        let outputPortMaxLength = 0;

        for (const portName of view.nameToPort.keys()) {
            const port = view.nameToPort.get(portName);
            if (port.direction === 'input') {
                inputPortMaxLength = Math.max(portName.length, inputPortMaxLength);
            } else {
                outputPortMaxLength = Math.max(portName.length, outputPortMaxLength);
            }
        }

        for (const portName of view.nameToPort.keys()) {
            const port = view.nameToPort.get(portName);
            if (port.direction === 'input') {
                inputCount ++;
                const offsetY = inputCount * LAYOUT_CONSTANT.PORT_TOP_MARGIN + 
                                (inputCount - 1) * LAYOUT_CONSTANT.CELL_PORT_HEIGHT +
                                LAYOUT_CONSTANT.INSTANCE_TITLE_HEIGHT;
                
                portnames.push({
                    name: portName,
                    x: 5,
                    y: offsetY,
                    align: 'left'
                });
            } else {
                outputCount ++;
                const offsetY = outputCount * LAYOUT_CONSTANT.PORT_TOP_MARGIN +
                                (outputCount - 1) * LAYOUT_CONSTANT.CELL_PORT_HEIGHT +
                                LAYOUT_CONSTANT.INSTANCE_TITLE_HEIGHT;

                portnames.push({
                    name: portName,
                    x: instanceWidth - 5,
                    y: offsetY,
                    align: 'end'
                });
            }
        }

        this.data.push({
            id: node.id,
            type: node.renderName,  // 例化模块的模块名字
            x: node.x,
            y: node.y,
            name: node.name,
            width: node.width,
            height: node.height,
            fill: 'var(--main-dark-color)',
            text: node.renderName,
            portnames,
            rx: 3,
            ry: 3,
            expandText: '📌',
            _self: node
        });
    }

    render() {
        const data = this.data;
        const rootRender = this.rootRender;
        const id2manager = this.id2manager;
        const _this = this;

        let instantiationSelections = this.parentSelection.selectAll('g.instance')
            .data(data)
            .enter()
            .append('g')
            .attr('class', 'instance')
            .attr("transform", d => `translate(${d.x}, ${d.y})`);
        
        let instances = instantiationSelections.append('rect')
            .attr('y', LAYOUT_CONSTANT.INSTANCE_TITLE_HEIGHT)
            .attr('width', data => data.width)
            .attr('height', data => data.height - LAYOUT_CONSTANT.INSTANCE_TITLE_HEIGHT)
            .attr('fill', d => {
                const children = d._self.children || [];
                return children.length > 0 ? 'rgba(203, 129, 218, 0.1)' : d.fill;
            });
        

        let texts = instantiationSelections.append('text')
            .attr('x', 0)
            .attr('y', LAYOUT_CONSTANT.INSTANCE_TITLE_HEIGHT - 8)
            .attr('dominant-baseline', 'middle')
            .attr('text-anchor', 'left')
            .attr('fill', 'var(--main-dark-color)')
            .attr('font-size', '0')
            .attr('font-size', '12px')
            .attr('class', 'port-caption')
            .text(data => data.name + ' ' + data.expandText)
            .each(function(data) {
                const text = d3.select(this);
                const bbox = text.node().getBBox();                
                instantiationSelections.insert('rect', 'text')
                    .attr('x', -1)
                    .attr('y', LAYOUT_CONSTANT.INSTANCE_TITLE_HEIGHT - bbox.height - 5)
                    .attr('width', bbox.width + 10)
                    .attr('height', bbox.height + 5)
                    .attr('fill', 'var(--main-color)');
            })
            .on('click', function(event) {
                const selection = d3.select(this);
                const d = selection.datum();
                const children = d._self.children || [];

                if (children.length > 0) {
                    rootRender.collapseInstance(selection);
                } else {
                    rootRender.expandInstance(selection);
                }
            });


        // 获取 port 相关的信息
        instantiationSelections.each(function(node) {
            d3.select(this)  // 选择当前节点
                .selectAll('.port-name')  // 选择所有端口名称元素
                .data(node.portnames)     // 绑定端口名称数据
                .enter()
                .append('text')           // 创建文本元素
                .attr('class', 'port-name')
                .attr('x', d => d.x)      // 使用端口的相对 x 坐标
                .attr('y', d => d.y)      // 使用端口的相对 y 坐标
                .attr('dominant-baseline', 'middle')  // 文本垂直居中
                .attr('text-anchor', d => d.align)
                .attr('fill', 'var(--foreground)')    // 文本颜色
                .attr('font-size', '10px')            // 设置字体大小
                .text(d => d.name);                   // 设置文本内容
        });

        if (globalSetting.renderAnimation) {
            instances.transition()
                .duration(1000)
                .attr('stroke', 'var(--main-color)')
                .attr('stroke-width', 2)
        } else {
            instances.attr('stroke', 'var(--main-color)')
                .attr('stroke-width', 2)
        }

        instantiationSelections
            .attr('class', 'grab')
            .each(function (data) {                    
                const portSelection = d3.select(this);
                const manager = _this.createDataManager(portSelection, data);

                // TODO: 实现拖拽
                // registerDragEvent(manager, rootRender);       
            });
            
        this.selections = instantiationSelections;
        return instantiationSelections;
    }

    /**
     * 
     * @param {d3.Selection} selection 
     * @param {BasicD3DataItem} data 
     * @returns {BasicD3ManagmentItem}
     */
    createDataManager(selection, data) {
        const id2manager = this.id2manager;
        // 创建拖拽上下文
        const dragContext = {
            neighbors: [],
            elkGraph: {
                // elk 是无状态的，id 取什么名字都行
                id: 'root',
                children: [],
                edges: [],
                layoutOptions: {}
            }
        }
        const managerItem = {
            data,
            selection,
            type: 'cell',
            dragContext
        };

        if (!id2manager.has(data.id)) {
            id2manager.set(data.id, []);
        }

        id2manager.get(data.id).push(managerItem);
        return managerItem;
    }
}